//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                          //
//                                               WiFi Paddle Boat                                           //
//                                                                                                          //
//                                 Copyright (c) 2018 By Zumwalt Properties, LLC.                           //
//                                                                                                          //
//                                             All Rights Reserved.                                         //
//                                                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Operates in AP mode, default ssid and password is "WiFiPaddleBoat".
//
// Requires the following electronics:
//
//  1) 1 Heltec WiFi Kit 32 with Oled display.
//  2) 1 Timesetl L298N Motor Drive Controller.
//  3) 2 300 RPM 6 VDC Mini Gear Motors.
//

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Includes.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include                              <WiFi.h>                              // for wifi
#include                              <ESP32WebServer.h>                    // for web server
#include                              <U8g2lib.h>                           // see https://github.com/olikraus/u8g2/wiki/u8g2reference
#include                              <WebSocketServer.h>                   // for the socket server (data transmission)
#include                              "freertos/FreeRTOS.h"                 // for tasks
#include                              "freertos/task.h"                     // for tasks

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Copyright.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////
     
  // Copyright start.
  
  const char*                         chCopyrightStart = "WiFi Paddle Boat, Copyright (c) 2018 by Zumwalt Properties, LLC.  All Rights Reserved";
  
//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Constants.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

  // Motor speed.

      #define                         MOTOR_MAX_SPEED   200                 // max motor speed (PWM)

  // Stop sign.

      // Sign width and height (must be a multilple of 6 and 2).
      
      #define                         SIGN_WIDTH        36
      #define                         SIGN_HEIGHT       36

      // Sign point outline (POINT1X and POINT1Y are the upper left hand corner of the sign,
      // the remaining points proceed clockwise).
      
      #define                         POINT1X           ((DISPLAY_WIDTH / 2) - (SIGN_WIDTH / 6))
      #define                         POINT1Y           ((DISPLAY_HEIGHT / 2) - (SIGN_HEIGHT / 2))
      #define                         POINT2X           ((DISPLAY_WIDTH / 2) + (SIGN_WIDTH / 6))
      #define                         POINT2Y           POINT1Y
      #define                         POINT3X           ((DISPLAY_WIDTH / 2) + (SIGN_WIDTH / 2))
      #define                         POINT3Y           ((DISPLAY_HEIGHT / 2) - (SIGN_HEIGHT / 6))
      #define                         POINT4X           POINT3X
      #define                         POINT4Y           ((DISPLAY_HEIGHT / 2) + (SIGN_HEIGHT / 6))
      #define                         POINT5X           POINT2X
      #define                         POINT5Y           ((DISPLAY_HEIGHT / 2) + (SIGN_HEIGHT / 2))
      #define                         POINT6X           POINT1X
      #define                         POINT6Y           POINT5Y
      #define                         POINT7X           ((DISPLAY_WIDTH / 2) - (SIGN_WIDTH / 2))
      #define                         POINT7Y           POINT4Y
      #define                         POINT8X           POINT7X
      #define                         POINT8Y           POINT3Y
      
  // WiFi Kit 32 Oled.

      #define                         DISPLAY_HEIGHT    64                  // display height in pixels
      #define                         DISPLAY_WIDTH     128                 // display width in pixels
      #define                         FONT_ONE_HEIGHT   8                   // font one height in pixels

  // WiFi Kit 32 to LN298 Motor Speed Controller pins and connections.

      // The port (left) motor is connected to motor speed controller OUT4 ("+") and OUT3 ("-"),
      // and is controlled by the following pins:
      
      #define                         EN_A              26                  // pin 26 on the WiFi Kit 32 is connected to the ENA pin on the motor speed controller
      #define                         IN_1              12                  // pin 12 on the WiFi Kit 32 is connected to the IN1 pin on the motor speed controller
      #define                         IN_2              13                  // pin 13 on the WiFi Kit 32 is connected to the IN2 pin on the motor speed controller
      
      // The starboard (right) motor is connected to motor speed controller OUT1 ("+") and OUT2 ("-")
      // and is controlled by the following pins:
      
      #define                         IN_3              27                  // pin 27 on the WiFi Kit 32 is connected to the IN3 pin on the motor speed controller
      #define                         IN_4              14                  // pin 14 on the WiFi Kit 32 is connected to the IN4 pin on the motor speed controller
      #define                         EN_B              25                  // pin 25 on the WiFi Kit 32 is connected to the ENB pin on the motor speed controller

      #define                         PWM_FREQUENCY     50                  // pwm frequency
      
  // HTTP and Socket port numbers.

      #define                         PORT_HTTP         80                  // port number for http
      #define                         PORT_SOCKET       2205                // port number for socket
      
//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Variables.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

    // Acccess point ssid.
    
      IPAddress                       IpAddress(192,168,20,20);               // access point ip address
      IPAddress                       IpGateway(192,168,20,1);                // access point gateway
      IPAddress                       IpSubnet(255,255,255,0);                // access point subnet mask
      String                          sPassword = "WiFiPaddleBoat";           // router and access point password
      String                          sSsidAccessPoint = "WiFiPaddleBoat";    // access point ssid
      
    //  Clients.

      int                             nClients = 0;                           // client socket count
      int                             nClientDisconnects = 0;                 // client disconnects

    // Graphics.

      U8G2_SSD1306_128X64_NONAME_F_HW_I2C u8g2(U8G2_R0, 16, 15, 4);           // OLED graphics

    // Motors.

    int                               nPortMotorPower = 0;                    // port motor power (0 through 100%)
    int                               nSpeed = 0;                             // motor speed (-100 through 100%)
    int                               nStarboardMotorPower = 0;               // starboard motor power (0 through 100%)
    
    // Servers.
      
      ESP32WebServer                  serverHTML(PORT_HTTP);                  // html server;
      WiFiServer                      serverSocket = (PORT_SOCKET);           // socket server
    
//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//  WiFi Paddle Boat web page.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

      // The WiFi Paddle Boat web page string consists of both Javascript (for touch control, touch control graphics
      // and data transmission), and html (for the web page).
      
      String        WiFiPaddleBoatWebPage=

          //////////////////////////////////////////////////////////////////////////////////////////////////////////////
          //
          // Javascript content of web page string.
          //
          //////////////////////////////////////////////////////////////////////////////////////////////////////////////
          
          "<script type='text/javascript'>\n"

              //
              // Variables.
              //
              
                  "var  canvasWiFiPaddleBoat;\n"
                  "var  canvasWiFiPaddleBoatHeight = 0;\n"
                  "var  canvasWiFiPaddleBoatWidth = 0;\n"
                  "var  contextWiFiPaddleBoat;\n"
                  "var  nWiFiPaddleBoatDelay = 0;\n"
                  "var  nWiFiPaddleBoatPort =" +  String(PORT_SOCKET) + ";\n"
                  "var  nWiFiPaddleBoatX = 0;\n"
                  "var  nWiFiPaddleBoatXN1 = -1;\n"
                  "var  nWiFiPaddleBoatY = 0;\n"
                  "var  nWiFiPaddleBoatYN1 = -1;\n"
                  "var  touchlistWiFiPaddleBoat = [];\n"
                  "var  wsWiFiPaddleBoat = new WebSocket('ws://' + location.hostname + ':' + nWiFiPaddleBoatPort + '/');\n"

              //
              // Utilities.
              //

                  //
                  // sendWiFiPaddleBoatXY()
                  // Send the nWiFiPaddleBoatX and nWiFiPaddleBoatX values to the server.
                  // Entry   : nothing
                  // Returns : nothing
                  //
                  // Notes   :
                  //
                  // The server (the WiFi Paddle Boat itself) is expecting the header "zpllc" immediately
                  // followed by the x and y touch coordinates in the range of 0 through 99, with
                  // leading zeros if necessary.
                  
                  "function sendWiFiPaddleBoatXY()\n"
                  "{\n"
                      // Convert x and y ints to strings.
                      
                      "var stringX = String(nWiFiPaddleBoatX);\n"
                      "var stringY = String(nWiFiPaddleBoatY);\n"
    
                      // The server is expecting 2 digit values for x and y (00 through 99),
                      // so add leading zero(s) to the strings as required.
                      
                      "while(stringX.length < 2)\n"
                          "stringX = '0' + stringX;\n"
                      
                      "while(stringY.length < 2)\n"
                          "stringY = '0' + stringY;\n"
    
                      // Combine the strings with the server validation header.
                      //
                      // Notes:
                      //
                      //  Again, the server is expecting an ascii string formatted as follows:
                      //
                      //      "zpllcXXYY"
                      //
                      //  where:
                      //
                      //    "zpllc": is a header for server validation of the message string,
                      //    "XX"   : is the x value,
                      //    "YY"   : is the y value.
                      //
                      
                      "stringX = 'zpllc' + stringX + stringY;\n"
                      
                      // Send the result to the server.
                      
                      "wsWiFiPaddleBoat.send(stringX);\n"
                      
                    "}\n" 

                  //
                  // drawWiFiPaddleBoatTouchImage(x, y)
                  // Draw the touch image on the canvas.
                  // Entry   : x and y location of where to draw the image
                  // Returns : nothing
                  //
                  
                  "function drawWiFiPaddleBoatTouchImage(x, y)\n"
                  "{\n"
                      // Start a drawing path.
                      
                      "contextWiFiPaddleBoat.beginPath();\n"

                      // Draw a blue dot.
                      
                      "contextWiFiPaddleBoat.arc(x, y, 40, 0, 2 * Math.PI, true);\n"
                      "contextWiFiPaddleBoat.fillStyle = 'rgba(0, 0, 200, 0.2)';\n"
                      "contextWiFiPaddleBoat.fill();\n"
                      "contextWiFiPaddleBoat.lineWidth = 2.0;\n"
                      "contextWiFiPaddleBoat.strokeStyle = 'rgba(0, 0, 200, 0.8)';\n"

                      // Present the drawing path.
                      
                      "contextWiFiPaddleBoat.stroke();\n"

                      // TODO:  saved for debugging.
                      
//                      "contextWiFiPaddleBoat.font = '30px Arial';\n"
//                      "contextWiFiPaddleBoat.fillText('ws://' + location.hostname + ':' + nWiFiPaddleBoatPort + '/', 10, 50);\n"

                  "}\n"

              //
              // onloadWiFiPaddleBoat()
              // Called by the html code when the web page is loaded.
              // Entry   : nothing
              // Returns : nothing
              //
              
              "function onloadWiFiPaddleBoat()\n"
              "{\n"
                  // Obtain the canvas from the html portion of the web page.
                  
                  "canvasWiFiPaddleBoat = document.getElementById('canvasWiFiPaddleBoat');\n"

                  // Obtain a drawing context for the canvas.
                  
                  "contextWiFiPaddleBoat = canvasWiFiPaddleBoat.getContext('2d');\n"

                  // Initialize the canvas.
                  
                  "canvasWiFiPaddleBoatWidth = window.innerWidth;\n"
                  "canvasWiFiPaddleBoatHeight = window.innerHeight;\n"
                  
                  "canvasWiFiPaddleBoat.style.width = canvasWiFiPaddleBoatWidth +'px';\n"
                  "canvasWiFiPaddleBoat.style.height = canvasWiFiPaddleBoatHeight + 'px';\n"
                  "canvasWiFiPaddleBoat.width = canvasWiFiPaddleBoatWidth;\n"
                  "canvasWiFiPaddleBoat.height = canvasWiFiPaddleBoatHeight;\n"

                  // Draw the initial touch image.
                  
                  "drawWiFiPaddleBoatTouchImage(window.innerWidth / 2, window.innerHeight / 2);\n"
                  
                  // Create a function for the "touchstart" event.
                  
                  "canvasWiFiPaddleBoat.addEventListener('touchstart', function(event)\n"
                  "{\n"
                      // Obtain list of touch points (we only use the first one in WiFi Paddle Boat).
                      
                      "touchlistWiFiPaddleBoat = event.touches;\n"

                      // Check for existance of a touch point.
                      
                      "if(touchlistWiFiPaddleBoat.length)\n"
                      "{\n"
                          // Have a touchpoint, update x.

                          "nWiFiPaddleBoatX = parseInt((100 * touchlistWiFiPaddleBoat[0].pageX) / canvasWiFiPaddleBoatWidth);\n"

                          "nWiFiPaddleBoatX = (nWiFiPaddleBoatX > 0) ? (nWiFiPaddleBoatX < 99) ? nWiFiPaddleBoatX : 99 : 0;\n"
                          
                          // Update y.
                          
                          "nWiFiPaddleBoatY = parseInt((100 * touchlistWiFiPaddleBoat[0].pageY) / canvasWiFiPaddleBoatHeight);\n"

                          "nWiFiPaddleBoatY = (nWiFiPaddleBoatY > 0) ? (nWiFiPaddleBoatY < 99) ? nWiFiPaddleBoatY : 99 : 0;\n"

                          // Send x and y position data to the server.

                              // Update n-1 values.
                              
                              "nWiFiPaddleBoatXN1 = nWiFiPaddleBoatX;\n"
                              "nWiFiPaddleBoatYN1 = nWiFiPaddleBoatY;\n"
                              "nWiFiPaddleBoatDelay = 8;\n"
                              
                              // Send x, y.
                              
                              "sendWiFiPaddleBoatXY();\n"
                          
                          // Update the touch image position.
                          
                          "contextWiFiPaddleBoat.clearRect(0, 0, canvasWiFiPaddleBoatWidth, canvasWiFiPaddleBoatHeight);\n"
                          "drawWiFiPaddleBoatTouchImage(touchlistWiFiPaddleBoat[0].pageX, touchlistWiFiPaddleBoat[0].pageY);\n"
                          
                      "}\n"
                  "});\n"
                  
                  // Create a function for the "touchmove" event.
                  
                  "canvasWiFiPaddleBoat.addEventListener('touchmove', function(event)\n"
                  "{\n"
                      // No scrolling during touch moves.
                      
                      "event.preventDefault();\n"

                      // Obtain list of touch points.
                      
                      "touchlistWiFiPaddleBoat = event.touches;\n"

                      // Check for existance of a touch point.
                      
                      "if(touchlistWiFiPaddleBoat.length)\n"
                      "{\n"
                          // Have a touchpoint, update x.

                          "nWiFiPaddleBoatX = parseInt((100 * touchlistWiFiPaddleBoat[0].pageX) / canvasWiFiPaddleBoatWidth);\n"

                          "nWiFiPaddleBoatX = (nWiFiPaddleBoatX > 0) ? (nWiFiPaddleBoatX < 99) ? nWiFiPaddleBoatX : 99 : 0;\n"
                          
                          // Update y.
                          
                          "nWiFiPaddleBoatY = parseInt((100 * touchlistWiFiPaddleBoat[0].pageY) / canvasWiFiPaddleBoatHeight);\n"

                          "nWiFiPaddleBoatY = (nWiFiPaddleBoatY > 0) ? (nWiFiPaddleBoatY < 99) ? nWiFiPaddleBoatY : 99 : 0;\n"

                          // Send x and y position data to the server.
                          //
                          // Notes:
                          //
                          //  In this application, there is no need to continuously send data to the server,
                          // so nWiFiPaddleBoatDelay is incorporated to send data to the server every 16 passes if
                          // either the x or y values have changed.

                          "if(((nWiFiPaddleBoatXN1 != nWiFiPaddleBoatX) || (nWiFiPaddleBoatYN1 != nWiFiPaddleBoatY)) && ((nWiFiPaddleBoatDelay --) <= 0))\n"
                          "{\n"
                                // x or y has changed and delay is up, update n-1 values and reset delay.
                                
                                "nWiFiPaddleBoatXN1 = nWiFiPaddleBoatX;\n"
                                "nWiFiPaddleBoatYN1 = nWiFiPaddleBoatY;\n"
                                "nWiFiPaddleBoatDelay = 8;\n"
                                
                                // Send x, y.
                                
                                "sendWiFiPaddleBoatXY();\n"
                          "}\n"
                          
                          // Update the touch image position.
                          
                          "contextWiFiPaddleBoat.clearRect(0, 0, canvasWiFiPaddleBoatWidth, canvasWiFiPaddleBoatHeight);\n"
                          "drawWiFiPaddleBoatTouchImage(touchlistWiFiPaddleBoat[0].pageX, touchlistWiFiPaddleBoat[0].pageY);\n"
                          
                      "}\n"
                  "});\n"

                   // Create a function for the "touchend" event.
                  
                  "canvasWiFiPaddleBoat.addEventListener('touchend', function()\n"
                  "{\n"
                      // At the end of the touch, return to center thus stopping WiFi Paddle Boat.

                      "nWiFiPaddleBoatX = 50;\n"
                      "nWiFiPaddleBoatY = 50;\n"

                      // Send x, y.
                      
                      "sendWiFiPaddleBoatXY();"
                      
                      // Draw the touch image.
                      
                      "contextWiFiPaddleBoat.clearRect(0, 0, canvasWiFiPaddleBoatWidth, canvasWiFiPaddleBoatHeight);\n"
                      "drawWiFiPaddleBoatTouchImage(canvasWiFiPaddleBoatWidth / 2, canvasWiFiPaddleBoatHeight / 2);\n"
                      
                   "});\n"
              "};\n"
               
          "</script>\n"
          
          //////////////////////////////////////////////////////////////////////////////////////////////////////////////
          //
          // HTML content of web page.
          //
          //////////////////////////////////////////////////////////////////////////////////////////////////////////////

          "<!DOCTYPE html>\n"
          
          "<html>\n"
          
              "<head>\n"
          
                  "<title>WiFi Paddle Boat</title>\n"
                  "<meta name='viewport' content='width=device-width, user-scalable=no'>\n"
          
              "<style type='text/css'>\n"
              
                  "body { margin: 0px; overflow: hidden; }\n"
                  "canvas { border: none; }\n"
                  
              "</style>\n"
          
              "</head>\n"
              
              "<body onload='onloadWiFiPaddleBoat()'>\n"

                  "<canvas id='canvasWiFiPaddleBoat' width='300' height='300' style='top:0px; left:0px; width: 300px; height: 300px;'></canvas>\n"
             
              "</body>\n"
          
          "</html>"
      ;

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Setup.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

void setup() 
{

  // Serial.
  
  Serial.begin(115200);
  while(!Serial){};
  
  Serial.printf("setup()...\n");

  // Wifi.

    // Wifi to access point mode.

    Serial.printf("  wifi mode ap.\n");
    WiFi.mode(WIFI_AP);

    // Set the access point ip address, gateway and subnet.
    
    Serial.printf("  wifi ap ip address, gateway and subnet.\n");
    WiFi.softAPConfig(IpAddress, IpGateway, IpSubnet);

    // Set the access point ssid and password.
    
    Serial.printf("  wifi ap ssid and password.\n");
    WiFi.softAP(sSsidAccessPoint.c_str(), sPassword.c_str());

  // OLED graphics.
  
  Serial.printf("  initialize OLED graphics.\n");
  
  u8g2.begin();
  u8g2.setFont(u8g2_font_6x10_tr);
  u8g2.setFontPosTop();
  u8g2.setDrawColor(1);
 
  // Display title.

  Serial.printf("  display title.\n");
  
  char  chBuffer[81];
  u8g2.clearBuffer();
  sprintf(chBuffer, "%s", "WiFi Paddle Boat");
  u8g2.drawStr(64 - (u8g2.getStrWidth(chBuffer) / 2), (DISPLAY_HEIGHT / 2) - (FONT_ONE_HEIGHT / 2), chBuffer);
  u8g2.sendBuffer();

  // Motor speed controller interface pins.

    Serial.printf("  initialize motor control.\n");
    
    // Port motor.
    
    pinMode(EN_A, OUTPUT);
    pinMode(IN_1, OUTPUT);
    pinMode(IN_2, OUTPUT);
    digitalWrite(EN_A, 0);
    ledcSetup(0, PWM_FREQUENCY, 8);
    ledcAttachPin(EN_A, 0);

    // Starboard motor.
    
    pinMode(EN_B, OUTPUT);
    pinMode(IN_3, OUTPUT);
    pinMode(IN_4, OUTPUT);
    digitalWrite(EN_B, 0);
    ledcSetup(1, PWM_FREQUENCY, 8);
    ledcAttachPin(EN_B, 1);

  // Delay.

  Serial.printf("  delay(4000), display copyright by, set up html server.\n");
  
  delay(4000);
  
  // Display copyright by.
  
  u8g2.clearBuffer();
  sprintf(chBuffer, "%s", "Copyright 2018");
  u8g2.drawStr(64 - (u8g2.getStrWidth(chBuffer) / 2), (DISPLAY_HEIGHT / 2) - FONT_ONE_HEIGHT, chBuffer);
  sprintf(chBuffer, "%s", "by");
  u8g2.drawStr(64 - (u8g2.getStrWidth(chBuffer) / 2), (DISPLAY_HEIGHT / 2) + FONT_ONE_HEIGHT, chBuffer);
  u8g2.sendBuffer();

  // Server.
  
      // Root is the webpage.
      //
      // "/" is sent from the http client and is a request for a complete webpage refresh.
      
        Serial.printf("  serverHTML.on().\n");
  
        serverHTML.on("/", []()
        {
          serverHTML.send(200, "text/html", WiFiPaddleBoatWebPage);
        });

      // Start the server.
      
        Serial.printf("  serverHTML.begin().\n");
  
        serverHTML.begin();
  
      // Start the socket.
        
        Serial.printf("  serverSocket.begin().\n");
  
        serverSocket.begin();

  // Delay.

  Serial.printf("  delay(4000), display copyright holder.\n");
  
  delay(4000);
  
  // Display copyright holder.
  
  u8g2.clearBuffer();
  sprintf(chBuffer, "%s", "Zumwalt Properties,");
  u8g2.drawStr(64 - (u8g2.getStrWidth(chBuffer) / 2), (DISPLAY_HEIGHT / 2) - FONT_ONE_HEIGHT, chBuffer);
  sprintf(chBuffer, "%s", "LLC.");
  u8g2.drawStr(64 - (u8g2.getStrWidth(chBuffer) / 2), (DISPLAY_HEIGHT / 2) + FONT_ONE_HEIGHT, chBuffer);
  u8g2.sendBuffer();

  // Delay.
  
  Serial.printf("  delay(4000), display all rights reserved.\n");
  
  delay(4000);

  // Display All Rights Reserved.

  u8g2.clearBuffer();
  sprintf(chBuffer, "%s", "All Rights Reserved.");
  u8g2.drawStr(64 - (u8g2.getStrWidth(chBuffer) / 2), (DISPLAY_HEIGHT / 2) - (FONT_ONE_HEIGHT / 2), chBuffer);
  u8g2.sendBuffer();

  // Delay.
  
  Serial.printf("  delay(4000), display ap and ip.\n");
  
  delay(4000);

  // Display ready.

  u8g2.clearBuffer();
  sprintf(chBuffer, "%s%s", "AP = ", sSsidAccessPoint.c_str());
  u8g2.drawStr(64 - (u8g2.getStrWidth(chBuffer) / 2), (DISPLAY_HEIGHT / 2) - FONT_ONE_HEIGHT, chBuffer);
  sprintf(chBuffer, "IP = %d.%d.%d.%d", IpAddress[0], IpAddress[1], IpAddress[2], IpAddress[3]);
  u8g2.drawStr(64 - (u8g2.getStrWidth(chBuffer) / 2), (DISPLAY_HEIGHT / 2) + FONT_ONE_HEIGHT, chBuffer);
  u8g2.sendBuffer();

  // Setup complete.
  
  Serial.printf("setup() complete, good luck!\n");
}
 
//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Loop.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

void loop() 
{    
    // Variables.

    static int   nStopSign = 0;
    
    // Handle html.
    
    serverHTML.handleClient();

    // Handle sockets.
  
      // Check for a new client connection.
      
      WiFiClient NewClient = serverSocket.available();
      if(NewClient)
      {
          if(NewClient.connected())
          {
              // Client connection, pass NewClient to HandleClientSocket for communication.  
      
              xTaskCreate(& WiFiPaddleBoatSocketEvent, "WiFiPaddleBoatSocketEvent", 2048, (void *) & NewClient, 1, NULL);
          }
      }
      
      // Update the display.

        if(nClients)
        {
            // Compute the displayed motor power levels.
          
              // Port.
              
              int nPortMotorPowerDisplayed = (DISPLAY_HEIGHT * nPortMotorPower * abs(nSpeed)) / (100 * 100);
          
              // Starboard.
              
              int nStarboardMotorPowerDisplayed = (DISPLAY_HEIGHT * nStarboardMotorPower * abs(nSpeed)) / (100 * 100);
          
            // Clear the display buffer.
            
            u8g2.clearBuffer();
          
            // Draw the port motor power.
            
            u8g2.drawBox(0, 
                         (DISPLAY_HEIGHT - 1) - nPortMotorPowerDisplayed,
                         16, 
                         nPortMotorPowerDisplayed);
          
             // Draw the starboard motor power.
             
            u8g2.drawBox((DISPLAY_WIDTH - 1) - 16, 
                         (DISPLAY_HEIGHT - 1) - nStarboardMotorPowerDisplayed,
                         16,
                         nStarboardMotorPowerDisplayed);
          
             // Draw the direction arrow.
          
             if(nPortMotorPowerDisplayed || nStarboardMotorPowerDisplayed)
             {
                if(nSpeed < 0)
                {
                    // Reverse.
          
                    u8g2.drawTriangle((DISPLAY_WIDTH / 2) - 16,
                                       DISPLAY_HEIGHT / 2,
                                       (DISPLAY_WIDTH / 2) + 16,
                                       DISPLAY_HEIGHT / 2,
                                       DISPLAY_WIDTH / 2,
                                       DISPLAY_HEIGHT - 1);
                                      
                }
                else
                {
                    // Forward.
                    
                    u8g2.drawTriangle((DISPLAY_WIDTH / 2) - 16,
                                       DISPLAY_HEIGHT / 2,
                                       (DISPLAY_WIDTH / 2) + 16,
                                       DISPLAY_HEIGHT / 2,
                                       DISPLAY_WIDTH / 2,
                                       0);
                }
             }
             else
             {
                    // Stopped, draw stop sign.
    
                    nStopSign += 1;
                    
                    if(nStopSign & 2)
                    {
                      u8g2.drawBox(POINT1X, POINT1Y, (POINT2X - POINT1X), (POINT6Y - POINT1Y));
                      u8g2.drawBox(POINT8X, POINT8Y, (POINT3X - POINT8X), (POINT7Y - POINT8Y));
                      u8g2.drawTriangle(POINT8X, POINT8Y, POINT1X, POINT1Y, POINT1X, POINT8Y);
                      u8g2.drawTriangle(POINT2X, POINT2Y, POINT3X, POINT3Y, POINT2X, POINT3Y);
                      u8g2.drawTriangle(POINT4X, POINT4Y, POINT5X, POINT5Y, POINT2X, POINT4Y);
                      u8g2.drawTriangle(POINT7X, POINT7Y, POINT6X, POINT6Y, POINT1X, POINT7Y);
                    }
             }
        }
        else
        {
            char      chBuffer[81];
            
            u8g2.clearBuffer();
            sprintf(chBuffer, "%s%s", "AP = ", sSsidAccessPoint.c_str());
            u8g2.drawStr(64 - (u8g2.getStrWidth(chBuffer) / 2), (DISPLAY_HEIGHT / 2) - FONT_ONE_HEIGHT, chBuffer);
            sprintf(chBuffer, "IP = %d.%d.%d.%d", IpAddress[0], IpAddress[1], IpAddress[2], IpAddress[3]);
            u8g2.drawStr(64 - (u8g2.getStrWidth(chBuffer) / 2), (DISPLAY_HEIGHT / 2) + FONT_ONE_HEIGHT, chBuffer);
            u8g2.sendBuffer();      
        }
          
  // Display the results.
  
  u8g2.sendBuffer();    

  // Give up some time.
  
  delay(100);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// WiFiPaddleBoatSocketEvent
// Process socket client events.
// Entry   : client number
//         : event type
//         : pointer to message received from client
//         : length of message received from client
// Returns : nothing
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////

void WiFiPaddleBoatSocketEvent(void * pClient)
{
    // Variables.
    
    WiFiClient        client =  * (WiFiClient *)pClient;
    IPAddress         ip;
    WebSocketServer   webserverSocket;

    // Count clients.

    nClients ++;
    
    // Check for client connected and client handshake.
   
    if(client.connected() && webserverSocket.handshake(client))
    {
        // Client is connected and client handshake successfull, obtain and display the client
        // ip address.
        
        ip = client.remoteIP();
        Serial.printf("WiFiPaddleBoat:  client connected, ip address = %d.%d.%d.%d, client count = %d, ram = %d.\n\n", 
                      ip[0], 
                      ip[1], 
                      ip[2], 
                      ip[3], 
                      nClients, 
                      esp_get_free_heap_size());

        // Process client communications.

        int       nCount = 0;
        
        while(client.connected())
        {
            // Process data received from client.
            
            String data = webserverSocket.getData();
            char  chBuffer[81];
            data.toCharArray(chBuffer, data.length() + 1);

            if(data.length() > 0)
            {
              // Data received, process it.
              
              double dData = strtod((const char *) & chBuffer[5], NULL);
              
              // Derive and limit the x component to 0 through 99.
              
              int  nX = (int)dData / 100;
              nX = (nX > 0) ? (nX < 99) ? nX : 99 : 0;
              
              // Derive and limit the y component to 0 through 99.
              
              int  nY = (int)dData % 100;
              nY = (nY > 0) ? (nY < 99) ? nY : 99 : 0;

              // Calculate the motor power (0 through 100%) and speed levels (-100 to 100%) via the x and y components.
              
                // Port.
                
                nPortMotorPower = (((nX < 50) ? nX : 50) * 100) / 50;
                nPortMotorPower = nPortMotorPower > 8 ? nPortMotorPower : 0;
                
                // Starboard.
                
                nStarboardMotorPower = (((nX > 50) ? (100 - nX) : 50) * 100) / 50;
                nStarboardMotorPower = nStarboardMotorPower > 8 ? nStarboardMotorPower : 0;
              
                // Speed.
                
                nSpeed = (50 - nY) * 2;
              
              // Set motor directions and speeds.

                // Direction.
                
                if(nSpeed < 0)
                {
                  // Reverse.

                    // Port motor direction.
                    
                    digitalWrite(IN_1, 0);
                    digitalWrite(IN_2, 1);
  
                    // Starboard moto direction.
                    
                    digitalWrite(IN_3, 0);
                    digitalWrite(IN_4, 1);
                }
                else
                {
                  // Forward.

                    // Port motor direction.
                    
                    digitalWrite(IN_1, 1);
                    digitalWrite(IN_2, 0);

                    // Starboard motor direction.
                    
                    digitalWrite(IN_3, 1);
                    digitalWrite(IN_4, 0);
                }

                // Speed.

                  // Port.
                  
                  ledcWrite(0, (MOTOR_MAX_SPEED * nPortMotorPower * abs(nSpeed)) / (100 * 100));

                  // Starboard.
                  
                  ledcWrite(1, (MOTOR_MAX_SPEED * nStarboardMotorPower * abs(nSpeed)) / (100 * 100));
              
             }

            // Send data to the client at about 2 hz.
 
            nCount ++;

            if((nCount % 50) == 0)
            {
                String    message = "zpllc";
                webserverSocket.sendData("zpllc");
            }

        // Give up some time.
        
        delay(10);
        }

    // Client disconnected.

      // Motors off.
      
      ledcWrite(0, 0);
      ledcWrite(1, 0);
      nPortMotorPower = nStarboardMotorPower = nSpeed = 0;

      // Stop client.
      
      client.stop();
    }

  // Terminate task.

  nClients --;
  nClientDisconnects ++;
  
  Serial.printf("WiFiPaddleBoat:  client disconnected, ip address = %d.%d.%d.%d, client count = %d, ram = %d.\n\n", ip[0], ip[1], ip[2], ip[3], nClients, esp_get_free_heap_size());
  
  vTaskDelete(NULL);  
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Copyright.
//
//////////////////////////////////////////////////////////////////////////////////////////////////////////////
     
  // Copyright end.
  
  const char*                         chCopyrightEnd = "WiFi Paddle Boat, Copyright (c) 2018 by Zumwalt Properties, LLC.  All Rights Reserved";
  

